import re
import typing
import uuid

primitives = typing.Union[int, str, bool, float]


def generate_uri(graph_base, username):
    """

    Args:
        graph_base ():
        username ():

    Returns:

    """
    return "{}{}".format(graph_base, username)


def generate_new_id():
    return str(uuid.uuid4())


def convert_snake(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()



#!/usr/bin/env python

# gquery.py: functions that deal with / transform SPARQL queries in grlc
# This code is from grlc
# grlc, the git repository linked data API constructor, automatically builds Web APIs using SPARQL queries stored
# in git repositories. http://grlc.io/

import json
from yaml import load, dump, parser, scanner
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper
from rdflib.plugins.sparql.parser import Query, UpdateUnit
from rdflib.plugins.sparql.processor import translateQuery
from pyparsing import ParseException
import logging
from pprint import pformat
import traceback
import re
import requests

# grlc modules
from obasparql import static

glogger = logging.getLogger(__name__)

XSD_PREFIX = 'PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>'


def _getDictWithKey(key, dict_list):
    """ Returns the first dictionary in dict_list which contains the given key"""
    for d in dict_list:
        if key in d:
            return d
    return None


def get_parameters(rq, variables, endpoint, query_metadata, auth=None):
    """
        ?_name The variable specifies the API mandatory parameter name. The value is incorporated in the query as plain literal.
        ?__name The parameter name is optional.
        ?_name_iri The variable is substituted with the parameter value as a IRI (also: number or literal).
        ?_name_en The parameter value is considered as literal with the language 'en' (e.g., en,it,es, etc.).
        ?_name_integer The parameter value is considered as literal and the XSD datatype 'integer' is added during substitution.
        ?_name_prefix_datatype The parameter value is considered as literal and the datatype 'prefix:datatype' is added during substitution. The prefix must be specified according to the SPARQL syntax.
    """

    internal_matcher = re.compile("__agg_\d+__")
    ## Basil-style variables
    variable_matcher = re.compile(
        "(?P<required>[_]{1,2})(?P<name>[^_]+)_?(?P<type>[a-zA-Z0-9]+)?_?(?P<userdefined>[a-zA-Z0-9]+)?.*$")

    parameters = {}
    for v in variables:
        if internal_matcher.match(v):
            continue

        match = variable_matcher.match(v)
        # TODO: currently only one parameter per triple pattern is supported
        if match:
            vname = match.group('name')
            vrequired = True if match.group('required') == '_' else False
            vtype = 'string'
            # All these can be None
            vcodes = get_enumeration(rq, vname, endpoint, query_metadata, auth)
            vdefault = get_defaults(rq, vname, query_metadata)
            vlang = None
            vdatatype = None
            vformat = None

            mtype = match.group('type')
            muserdefined = match.group('userdefined')

            if mtype in ['number', 'literal', 'string']:
                vtype = mtype
            elif mtype in ['iri']:  # TODO: proper form validation of input parameter uris
                vtype = 'string'
                vformat = 'iri'
            elif mtype:
                vtype = 'string'

                if mtype in static.XSD_DATATYPES:
                    vdatatype = 'xsd:{}'.format(mtype)
                elif len(mtype) == 2:
                    vlang = mtype
                elif muserdefined:
                    vdatatype = '{}:{}'.format(mtype, muserdefined)

            parameters[vname] = {
                'original': '?{}'.format(v),
                'required': vrequired,
                'name': vname,
                'type': vtype
            }

            # Possibly None parameter attributes
            if vcodes is not None:
                parameters[vname]['enum'] = sorted(vcodes)
            if vlang is not None:
                parameters[vname]['lang'] = vlang
            if vdatatype is not None:
                parameters[vname]['datatype'] = vdatatype
            if vformat is not None:
                parameters[vname]['format'] = vformat
            if vdefault is not None:
                parameters[vname]['default'] = vdefault

            glogger.info('Finished parsing the following parameters: {}'.format(parameters))

    return parameters


def get_defaults(rq, v, metadata):
    """
    Returns the default value for a parameter or None
    """
    glogger.debug("Metadata with defaults: {}".format(metadata))
    if 'defaults' not in metadata:
        return None
    defaultsDict = _getDictWithKey(v, metadata['defaults'])
    if defaultsDict:
        return defaultsDict[v]
    return None


def get_enumeration(rq, v, endpoint, metadata={}, auth=None):
    """
    Returns a list of enumerated values for variable 'v' in query 'rq'
    """
    # glogger.debug("Metadata before processing enums: {}".format(metadata))
    # We only fire the enum filling queries if indicated by the query metadata
    if 'enumerate' not in metadata:
        return None
    enumDict = _getDictWithKey(v, metadata['enumerate'])
    if enumDict:
        return enumDict[v]
    if v in metadata['enumerate']:
        return get_enumeration_sparql(rq, v, endpoint, auth)
    return None


def get_enumeration_sparql(rq, v, endpoint, auth=None):
    """
    Returns a list of enumerated values for variable 'v' in query 'rq'
    """
    glogger.info('Retrieving enumeration for variable {}'.format(v))
    vcodes = []
    # tpattern_matcher = re.compile(".*(FROM\s+)?(?P<gnames>.*)\s+WHERE.*[\.\{][\n\t\s]*(?P<tpattern>.*\?" + re.escape(v) + ".*?\.).*", flags=re.DOTALL)
    # tpattern_matcher = re.compile(".*?((FROM\s*)(?P<gnames>(\<.*\>)+))?\s*WHERE\s*\{(?P<tpattern>.*)\}.*", flags=re.DOTALL)

    # WHERE is optional too!!
    tpattern_matcher = re.compile(".*?(FROM\s*(?P<gnames>\<.*\>+))?\s*(WHERE\s*)?\{(?P<tpattern>.*)\}.*",
                                  flags=re.DOTALL)

    glogger.debug(rq)
    tp_match = tpattern_matcher.match(rq)
    if tp_match:
        vtpattern = tp_match.group('tpattern')
        gnames = tp_match.group('gnames')
        glogger.debug("Detected graph names: {}".format(gnames))
        glogger.debug("Detected BGP: {}".format(vtpattern))
        glogger.debug("Matched triple pattern with parameter")
        if gnames:
            codes_subquery = re.sub("SELECT.*\{.*\}.*",
                                    "SELECT DISTINCT ?" + v + " FROM " + gnames + " WHERE { " + vtpattern + " }", rq,
                                    flags=re.DOTALL)
        else:
            codes_subquery = re.sub("SELECT.*\{.*\}.*",
                                    "SELECT DISTINCT ?" + v + " WHERE { " + vtpattern + " }", rq,
                                    flags=re.DOTALL)
        glogger.debug("Codes subquery: {}".format(codes_subquery))
        glogger.debug(endpoint)
        codes_json = requests.get(endpoint, params={'query': codes_subquery},
                                  headers={'Accept': static.mime_types['json'],
                                           'Authorization': 'token {}'.format(static.ACCESS_TOKEN)}, auth=auth).json()
        for code in codes_json['results']['bindings']:
            vcodes.append(list(code.values())[0]["value"])
    else:
        glogger.debug("No match between variable name and query.")

    return vcodes


def get_yaml_decorators(rq):
    """
    Returns the yaml decorator metadata only (this is needed by triple pattern fragments)
    """
    # glogger.debug('Guessing decorators for query {}'.format(rq))
    if not rq:
        return None

    if isinstance(rq, dict) and 'grlc' in rq:  # json query (sparql transformer)
        yaml_string = rq['grlc']
        query_string = rq
    else:  # classic query
        yaml_string = "\n".join([row.lstrip('#+') for row in rq.split('\n') if row.startswith('#+')])
        query_string = "\n".join([row for row in rq.split('\n') if not row.startswith('#+')])

    query_metadata = None
    if type(yaml_string) == dict:
        query_metadata = yaml_string
    elif type(yaml_string) == str:
        try:  # Invalid YAMLs will produce empty metadata
            query_metadata = load(yaml_string, Loader=Loader)
        except (parser.ParserError, scanner.ScannerError) as e:
            try:
                query_metadata = json.loads(yaml_string)
            except json.JSONDecodeError:
                glogger.warning("Query decorators could not be parsed; check your YAML syntax")

    # If there is no YAML string
    if query_metadata is None:
        query_metadata = {}
    query_metadata['query'] = query_string

    # glogger.debug("Parsed query decorators: {}".format(query_metadata))

    return query_metadata


def enable_custom_function_prefix(rq, prefix):
    if ' %s:' % prefix in rq or '(%s:' % prefix in rq and not 'PREFIX %s:' % prefix in rq:
        rq = 'PREFIX %s: <:%s>\n' % (prefix, prefix) + rq
    return rq


def get_metadata(rq, endpoint):
    """
    Returns the metadata 'exp' parsed from the raw query file 'rq'
    'exp' is one of: 'endpoint', 'tags', 'summary', 'request', 'pagination', 'enumerate'
    """
    query_metadata = get_yaml_decorators(rq)
    query_metadata['type'] = 'UNKNOWN'
    query_metadata['original_query'] = rq

    if isinstance(rq, dict):  # json query (sparql transformer)
        pass

    rq = enable_custom_function_prefix(rq, 'bif')
    rq = enable_custom_function_prefix(rq, 'sql')

    try:
        # THE PARSING
        # select, describe, construct, ask
        parsed_query = translateQuery(Query.parseString(rq, parseAll=True))
        query_metadata['type'] = parsed_query.algebra.name
        if query_metadata['type'] == 'SelectQuery':
            # Projection variables
            query_metadata['variables'] = parsed_query.algebra['PV']
            # Parameters
            query_metadata['parameters'] = get_parameters(rq, parsed_query.algebra['_vars'], endpoint, query_metadata)
        elif query_metadata['type'] == 'ConstructQuery':
            # Parameters
            query_metadata['parameters'] = get_parameters(rq, parsed_query.algebra['_vars'], endpoint, query_metadata)
        else:
            glogger.warning(
                "Query type {} is currently unsupported and no metadata was parsed!".format(query_metadata['type']))
    except ParseException as pe:
        glogger.warning(pe)
        glogger.warning("Could not parse regular SELECT, CONSTRUCT, DESCRIBE or ASK query")
        # glogger.warning(traceback.print_exc())

        # insert queries won't parse, so we regex
        # glogger.info("Trying to parse INSERT query")
        # if static.INSERT_PATTERN in rq:
        #     query_metadata['type'] = 'InsertQuery'
        #     query_metadata['parameters'] = [u'_g_iri']

        try:
            # update query
            glogger.info("Trying to parse UPDATE query")
            parsed_query = UpdateUnit.parseString(rq, parseAll=True)
            glogger.info(parsed_query)
            query_metadata['type'] = parsed_query[0]['request'][0].name
            if query_metadata['type'] == 'InsertData':
                query_metadata['parameters'] = {
                    'g': {'datatype': None, 'enum': [], 'lang': None, 'name': 'g', 'original': '?_g_iri',
                          'required': True, 'type': 'iri'},
                    'data': {'datatype': None, 'enum': [], 'lang': None, 'name': 'data', 'original': '?_data',
                             'required': True, 'type': 'literal'}}

            glogger.info("Update query parsed with {}".format(query_metadata['type']))
            # if query_metadata['type'] == 'InsertData':
            #     query_metadata['variables'] = parsed_query.algebra['PV']
        except:
            glogger.error("Could not parse query")
            glogger.error(query_metadata['query'])
            glogger.error(traceback.print_exc())
            pass

    glogger.debug("Finished parsing query of type {}".format(query_metadata['type']))
    glogger.debug("All parsed query metadata (from decorators and content): ")
    glogger.debug(pformat(query_metadata, indent=32))

    return query_metadata


def rewrite_query(query, parameters, get_args):
    glogger.debug("Query parameters")
    glogger.debug(parameters)
    requireXSD = False

    required_params = {}
    for k, v in parameters.items():
        if parameters[k]['required']:
            required_params[k] = v
    requiredParams = set(required_params.keys())
    providedParams = set(get_args.keys())
    glogger.debug("Required parameters: {} Request args: {}".format(requiredParams, providedParams))
    assert requiredParams.issubset(providedParams), 'Provided parameters do not cover the required parameters!'

    for pname, p in list(parameters.items()):
        # Get the parameter value from the GET request
        v = get_args.get(pname, None)
        # If the parameter has a value
        if not v:
            continue

        if isinstance(query, dict):  # json query (sparql transformer)
            if '$values' not in query:
                query['$values'] = {}
            values = query['$values']

            if not p['original'] in values:
                values[p['original']] = v
            elif isinstance(values[p['original']], list):
                values[p['original']].append(v)
            else:
                values[p['original']] = [values[p['original']], v]

            continue

        # IRI
        if p['type'] == 'iri':  # TODO: never reached anymore, since iris are now type=string with format=iri
            query = query.replace(p['original'], "{}{}{}".format('<', v, '>'))
        # A number (without a datatype)
        elif p['type'] == 'number':
            query = query.replace(p['original'], v)
        # Literals
        elif p['type'] == 'literal' or p['type'] == 'string':
            # If it's a iri
            if 'format' in p and p['format'] == 'iri':
                query = query.replace(p['original'], "{}{}{}".format('<', v, '>'))
            # If there is a language tag
            if 'lang' in p and p['lang']:
                query = query.replace(p['original'], "\"{}\"@{}".format(v, p['lang']))
            elif 'datatype' in p and p['datatype']:
                query = query.replace(p['original'], "\"{}\"^^{}".format(v, p['datatype']))
                if 'xsd' in p['datatype']:
                    requireXSD = True
            else:
                query = query.replace(p['original'], "\"{}\"".format(v))

    if isinstance(query, dict):  # json query (sparql transformer)
        pass

    if requireXSD and XSD_PREFIX not in query:
        query = query.replace('SELECT', XSD_PREFIX + '\n\nSELECT')

    glogger.debug("Query rewritten as: " + query)

    return query



from enum import Enum

mime_types = {
    'csv': 'text/csv; q=1.0, */*; q=0.1',
    'json': 'application/json; q=1.0, application/sparql-results+json; q=0.8, */*; q=0.1',
    'html': 'text/html; q=1.0, */*; q=0.1',
    'ttl': 'text/turtle'
}
QUERY_TYPE_GET_ALL = "get_all"
QUERY_TYPE_GET_ALL_USER = "get_all_user"
QUERY_TYPE_GET_ONE = "get_one"
QUERY_TYPE_GET_ONE_USER = "get_one_user"
QUERIES_TYPES = [QUERY_TYPE_GET_ALL, QUERY_TYPE_GET_ALL_USER, QUERY_TYPE_GET_ONE, QUERY_TYPE_GET_ONE_USER]

# PARAMETERS FROM THE API
ID_KEY = "id"
USERNAME_KEY = "username"
LABEL_KEY = "label"
PAGE_KEY = "page"
PER_PAGE_KEY = "per_page"

CUSTOM_QUERY_NAME = "custom_query_name"

CONTEXT_KEY = "@context"
CONTEXT_ID_KEY = "id"
CONTEXT_TYPE_KEY = "type"

SPARQL_QUERY_TYPE_VARIABLE = "type"
SPARQL_GRAPH_TYPE_VARIABLE = "g"
SPARQL_ID_TYPE_VARIABLE = "resource"

SKIP_ID_FRAMING_KEY = "skip_id_framing"

CONTEXT_FILE = "context.json"
CONTEXT_CLASS_FILE = "context_class.json"
CONTEXT_OVERWRITE_CLASS_FILE = "context_overwrite.json"

#Directory where the default queries are
DEFAULT_DIR = "_default_"

class Query(Enum):
    GET_ALL = ("get_all")


EMBED_OPTION = "@always"
JSONLD = 'json-ld'

XSD_DATATYPES = ["decimal", "float", "double", "integer", "positiveInteger", "negativeInteger", "nonPositiveInteger",
                 "nonNegativeInteger", "long", "int", "short", "byte", "unsignedLong", "unsignedInt", "unsignedShort",
                 "unsignedByte", "dateTime", "date", "gYearMonth", "gYear", "duration", "gMonthDay", "gDay", "gMonth",
                 "string", "normalizedString", "token", "language", "NMTOKEN", "NMTOKENS", "Name", "NCName", "ID",
                 "IDREFS", "ENTITY", "ENTITIES", "QName", "boolean", "hexBinary", "base64Binary", "anyURI", "notation"]



import json
import logging.config
import os
from pathlib import Path
from typing import Dict, Tuple
from starlette.exceptions import HTTPException

import validators
from pyld import jsonld
from rdflib import Graph
from obasparql.sparqlconnector import SPARQLConnector
from obasparql import gquery
from obasparql.static import CONTEXT_ID_KEY, CONTEXT_OVERWRITE_CLASS_FILE, CONTEXT_TYPE_KEY, CUSTOM_QUERY_NAME, DEFAULT_DIR, ID_KEY, PAGE_KEY, PER_PAGE_KEY, QUERY_TYPE_GET_ONE, QUERY_TYPE_GET_ONE_USER, QUERY_TYPE_GET_ALL, \
    CONTEXT_FILE, CONTEXT_CLASS_FILE, CONTEXT_KEY, SPARQL_QUERY_TYPE_VARIABLE, USERNAME_KEY, QUERY_TYPE_GET_ALL_USER, SPARQL_ID_TYPE_VARIABLE, SPARQL_GRAPH_TYPE_VARIABLE, SKIP_ID_FRAMING_KEY, EMBED_OPTION, JSONLD, XSD_DATATYPES
from obasparql.utils import generate_new_id, primitives, convert_snake

glogger = logging.getLogger("grlc")
logger = logging.getLogger('fastapi')


def remove_jsonld_key(tmp_context_class, key):
    """Remove json key from context

    Args:
        tmp_context_class (dict): the dict to remove the key from
        key (str): the key to remove
    """
    try:
        tmp_context_class.pop(key)
    except KeyError:
        logging.debug("The context file does not contains the id or type key")


class QueryManager:
    """Class to handle queries
    """

    def __init__(self,
                 endpoint: str,
                 named_graph_base: str,
                 uri_prefix: str,
                 queries_dir: str,
                 context_dir: str,
                 endpoint_username=None,
                 endpoint_password=None):
        """Constructor of the QueryManager class

        Args:
            endpoint (str): the endpoint server
            named_graph_base (str): the prefix or base of the name graphs
            uri_prefix (str): prefix for the URIs of new resource
            queries_dir (str): the directory where the queries are stored
            context_dir (str): the directory where the context are
            endpoint_username (str, optional): the username to access the endpoint. Defaults to None.
            endpoint_password ([type], optional): [description]. Defaults to None.

        Raises:
            e: [description]
            e: [description]
            e: [description]

        Returns:
            [type]: [description]
        """
        self.endpoint = endpoint
        self.endpoint_username = endpoint_username
        self.endpoint_password = endpoint_password
        self.update_endpoint = f'{self.endpoint}/update'
        self.query_endpoint = f'{self.endpoint}/query'
        self.named_graph_base = named_graph_base
        self.uri_prefix = uri_prefix
        self.sparql = SPARQLConnector(query_endpoint=self.endpoint,
                                      update_endpoint=self.update_endpoint,
                                      auth=(self.endpoint_username,
                                            self.endpoint_password),
                                      method="POST",
                                      returnFormat='json-ld')
        queries_dir = Path(queries_dir)
        context_dir = Path(context_dir)
        default_dir = queries_dir / DEFAULT_DIR

        # Obtain default queries
        queries = self.read_template(default_dir)
        setattr(self, "_default_", queries)

        # Overwrite default queries by class
        for owl_class in os.listdir(queries_dir):
            queries = self.read_template(queries_dir / owl_class)
            setattr(self, owl_class, self._default_)
            for key, _ in queries.items():
                k = getattr(self, owl_class)
                k[key] = queries[key]
        for query_name, query_sparql in queries.items():
            glogger.debug(query_name)
            glogger.debug(query_sparql)

        try:
            context_json = CONTEXT_FILE
            context_class_json = CONTEXT_CLASS_FILE
            temp_context = json.loads(
                self.read_context(context_dir / context_json))[CONTEXT_KEY]
            tmp_context_class = json.loads(
                self.read_context(context_dir /
                                  context_class_json))[CONTEXT_KEY]
        except FileNotFoundError:
            logging.error("The context file does not exists", exc_info=True)
            exit(1)

        try:
            context_overwrite_json = CONTEXT_OVERWRITE_CLASS_FILE
            context_overwrite_path = context_dir / context_overwrite_json
            self.context_overwrite = json.loads(
                self.read_context(context_overwrite_path))[CONTEXT_KEY]
        except FileNotFoundError:
            self.context_overwrite = {}

        remove_jsonld_key(tmp_context_class, CONTEXT_TYPE_KEY)
        remove_jsonld_key(tmp_context_class, CONTEXT_ID_KEY)

        self.context = temp_context.copy()
        self.convert_snake_dict(temp_context)
        self.context = {CONTEXT_KEY: self.context}
        self.class_context = tmp_context_class.copy()

    def get_resource(self, **kwargs):
        """
        Handle the GET Requests
        Parameters
        ----------
        kwargs:

        Returns
        -------

        """

        request_args: Dict[str, str] = {}
        if PAGE_KEY in kwargs:
            request_args[PAGE_KEY] = kwargs[PAGE_KEY]
        if PER_PAGE_KEY in kwargs:
            request_args[PER_PAGE_KEY] = kwargs[PER_PAGE_KEY]

        if CUSTOM_QUERY_NAME in kwargs:
            response = self.get_resource_custom(request_args=request_args,
                                                **kwargs)
        else:
            response = self.get_resource_not_custom(request_args=request_args,
                                                    **kwargs)
        return response

    def put_resource(self, id, user, body, rdf_type_uri, rdf_type_name=None, kls=None):
        """Handle a PUT method to update a resource

        Returns:
            dict: The response of the request as JSON format
        """
        resource_uri = self.build_instance_uri(id)
        body = body.dict()
        body["id"] = resource_uri
        try:
            username = user
        except Exception as err:
            logger.error("Missing username", exc_info=True)
            raise HTTPException(
                status_code=400, detail="Bad request: missing username") from err

        response = self.get_resource(
            id=id, username=username, rdf_type_uri=rdf_type_uri, rdf_type_name=rdf_type_name, kls=kls)
        # DELETE QUERY
        request_args_delete: Dict[str, str] = {
            "resource": resource_uri,
            "g": self.generate_graph(username),
            "delete_incoming_relations": False
        }
        self.run_query_delete(request_args_delete)

        # INSERT QUERY
        body_json = self.json_to_jsonld(body)
        prefixes, triples = self.build_query_insert(body_json)
        prefixes = '\n'.join(prefixes)
        triples = '\n'.join(triples)

        request_args: Dict[str, str] = {
            "prefixes": prefixes,
            "triples": triples,
            "g": self.generate_graph(username)
        }
        try:
            self.run_query_insert(request_args=request_args)
        except Exception as err:
            logger.error("Exception occurred", exc_info=True)
            raise HTTPException(status_code=500, detail=str(err)) from err

        # GET QUERY
        try:
            response = self.get_resource(
                id=id, username=username, rdf_type_uri=rdf_type_uri, rdf_type_name=rdf_type_name, kls=kls)
            return response
        except Exception as err:
            logger.error("Error while retrieving the resource", exc_info=True)
            raise HTTPException(
                status_code=500, detail="Error while retrieving the resource") from err

    def delete_resource(self,
                        id: str,
                        user: str,
                        rdf_type_uri: str = None,
                        rdf_type_name: str = None,
                        kls: str = None) -> dict:
        """Handle a DELETE method to delete a resource

        Args:
            id (str): the resource id
            user (str): the user who is deleting the resource
            rdf_type_uri (str, optional): The rdf type uri of the resource. Defaults to None.
            rdf_type_name (str, optional): The class name of the resource. Defaults to None.
            kls (str, optional): TODO: I don't remember. Defaults to None.

        Returns:
            dict: The response of the request as JSON format
        """
        resource_uri = self.build_instance_uri(id)
        request_args: Dict[str, str] = {
            "resource": resource_uri,
            "g": self.generate_graph(user),
            "delete_incoming_relations": True
        }
        # GET QUERY
        try:
            response = self.get_resource(
                id=resource_uri, username=user, rdf_type_uri=rdf_type_uri, rdf_type_name=rdf_type_name, kls=kls)
        except Exception as err:
            logger.error("Error while retrieving the resource", exc_info=True)
            raise HTTPException(
                status_code=500, detail="Error while retrieving the resource") from err

        if not response:
            logger.error("Error while retrieving the resource", exc_info=True)
            raise HTTPException(status_code=404, detail="Resource not found")

        self.run_query_delete(request_args)

    def post_resource(self,
                      user,
                      body,
                      rdf_type_uri,
                      rdf_type_name=None,
                      kls=None):
        """
        Post a resource and generate the id
        Args:
            username: named graph where to do the insert
            body: JSON to insert
            rdf_type_uri: RDF Class where to insert the target instance described in body.
        """
        if body.type and rdf_type_uri is not body.type:
            body.type.append(rdf_type_uri)
        else:
            body.type = [rdf_type_uri]
        body = body.dict()
        body[ID_KEY] = generate_new_id()
        self.traverse_obj(body, user)
        self.insert_all_resources(body, user)
        if '@context' in body:
            del body['@context']
        return body

    # Helper get methods
    def get_resource_custom(self, request_args, **kwargs):
        """
        Prepare request for custom queries
        : param request_args: contains the values to replaced in the query
        : param kwargs:
        : return:
        """
        query_type = kwargs[CUSTOM_QUERY_NAME]
        if ID_KEY in kwargs:
            return self.get_one_resource(request_args=request_args,
                                         query_type=query_type,
                                         **kwargs)
        else:
            if "label" in kwargs and kwargs["label"] is not None:
                request_args["label"] = kwargs["label"]
            return self.get_all_resource(request_args=request_args,
                                         query_type=query_type,
                                         **kwargs)

    def get_resource_not_custom(self, request_args, **kwargs):
        """
        Prepare request for not-custom queries

        : param request_args: contains the values to replaced in the query
        : param kwargs:
        : return:
        """
        if ID_KEY in kwargs:
            if USERNAME_KEY in kwargs:
                query_type = QUERY_TYPE_GET_ONE_USER
            else:
                query_type = QUERY_TYPE_GET_ONE
            return self.get_one_resource(request_args=request_args,
                                         query_type=query_type,
                                         **kwargs)

        elif ID_KEY not in kwargs:
            # TODO: Support label search
            # if LABEL_KEY in kwargs and kwargs[LABEL_KEY] is not None:
            # logging.warning("not supported")
            if USERNAME_KEY in kwargs:
                query_type = QUERY_TYPE_GET_ALL_USER
            elif USERNAME_KEY not in kwargs:
                query_type = QUERY_TYPE_GET_ALL
            return self.get_all_resource(request_args=request_args,
                                         query_type=query_type,
                                         **kwargs)

    def get_one_resource(self, request_args, query_type, **kwargs):
        """
        Handles a GET method to get one resource
        : param query_type: QUERY_TYPE_GET_ONE or QUERY_TYPE_GET_ONE_USER
        : param request_args:
        : param kwargs:
        : type kwargs:
        : return:
        : rtype:
        """
        owl_class_name, resource_type_uri, username = self.parse_request_arguments(
            **kwargs)
        request_args[SPARQL_ID_TYPE_VARIABLE] = self.build_instance_uri(
            kwargs[ID_KEY])
        request_args[SPARQL_GRAPH_TYPE_VARIABLE] = self.generate_graph(
            username)
        skip_id_framing = True if SKIP_ID_FRAMING_KEY in kwargs and kwargs[
            SKIP_ID_FRAMING_KEY] else False
        return self.run_query_get(query_directory=owl_class_name,
                                  owl_class_uri=resource_type_uri,
                                  query_type=query_type,
                                  request_args=request_args,
                                  skip_id_framing=skip_id_framing)

    def get_all_resource(self, request_args, query_type, **kwargs):
        """
        Handles a GET method to get all resource by rdf_type
        : param request_args: contains the values of the variables of the SPARQL query.
                             See SPARQL_QUERY_TYPE_VARIABLE and SPARQL_GRAPH_TYPE_VARIABLE
        : param query_type: QUERY_TYPE_GET_ALL or QUERY_TYPE_GET_ALL_USER
        : param kwargs:
        : type kwargs:
        : return:
        : rtype:
        """
        owl_class_name, resource_type_uri, username = self.parse_request_arguments(
            **kwargs)
        request_args[SPARQL_QUERY_TYPE_VARIABLE] = resource_type_uri
        request_args[SPARQL_GRAPH_TYPE_VARIABLE] = self.generate_graph(
            username)
        return self.run_query_get(query_directory=owl_class_name,
                                  owl_class_uri=resource_type_uri,
                                  query_type=query_type,
                                  request_args=request_args)

    # SPARQL AND JSON LD METHODS

    def run_query_get(self,
                      query_directory,
                      owl_class_uri,
                      query_type,
                      request_args=None,
                      skip_id_framing=False):
        """Generate the query, dispatch it to the SPARQL endpoint, frame it and return the response

        Args:
            query_directory (str): The directory where the query template is located
            owl_class_uri (str): The uri of the class 
            query_type (str): The type of query
            request_args (str, optional): The arguments of the query. Defaults to None.
            skip_id_framing (bool, optional): Indicates if the id framing must be skipped. Defaults to False.

        Raises:
            e: [description]

        Returns:
            [type]: [description]
        """
        query_template = getattr(self, query_directory)[query_type]
        if PAGE_KEY in request_args and PER_PAGE_KEY in request_args:
            request_args["offset"] = (request_args[PAGE_KEY] -
                                      1) * request_args[PER_PAGE_KEY]
        try:
            result = self.dispatch_sparql_query(
                raw_sparql_query=query_template, request_args=request_args)
        except Exception as err:
            logger.error("Unable to send query: %s", err)
            raise HTTPException(status_code=500, detail="Unable to send query") from err

        logger.debug("response: %s", result)
        if result is None:
            raise HTTPException(status_code=404, detail="Resource not found")

        try:
            if "resource" in request_args and not skip_id_framing:
                return self.frame_results(result, owl_class_uri,
                                          request_args["resource"])
            return self.frame_results(result, owl_class_uri)
        except Exception as err:
            raise HTTPException(
                status_code=500, detail="Unable to frame the results") from err

    def overwrite_endpoint_context(self, endpoint_context: dict):
        """Overwrite the endpoint context

        Args:
            endpoint_context (dict): The endpoint context to be overwrite
        """
        for key, value in self.context_overwrite.items():
            if key in endpoint_context:
                endpoint_context[key] = value

    def frame_results(self,
                      response: str,
                      owl_class_uri: str,
                      owl_resource_iri: str = None):
        """Frame the results of the query

        Args:
            response (str): The response from the endpoint. The format is JSON-LD
            owl_class_uri (str): The uri of the class
            owl_resource_iri (str, optional): The resource uri. Defaults to None.

        Returns:
            [type]: [description]
        """
        try:
            response_dict = json.loads(response)
        except Exception as exception:
            logger.error("json serialize failed", exc_info=True)
            raise exception

        if len(response_dict) == 0:
            # Fuseki sometimes return a json with graph and something not...
            if "@id" in response_dict and '@graph' not in response_dict:
                return []

        if '@context' in response_dict:
            response_context = response_dict['@context']
        else:
            response_context = {}

        if '@graph' in response_dict:
            response_graph = response_dict['@graph']
        else:
            response_graph = response_dict

        frame = {"@context": response_context, "@type": owl_class_uri}
        # if owl_resource_iri is set, the user is requesting a specific resource and we must add it to the frame
        if owl_resource_iri is not None:
            frame['@id'] = owl_resource_iri
        frame["@context"][ID_KEY] = "@id"

        # we must force that the type is a list
        frame["@context"]["type"] = {"@container": "@set", "@id": "@type"}
        # we must force that all the properties are lists
        for prop in frame["@context"].keys():
            if isinstance(frame["@context"][prop], dict):
                frame["@context"][prop]["@container"] = "@set"

        # TODO: I don't know why but the following line is needed to avoid the error:
        if 'id' in response_graph:
            del response_graph[ID_KEY]

        # Context (returned by the endpoint) does not contain information about the classes
        # so we must add it to the frame
        self.overwrite_endpoint_context(frame["@context"])
        frame["@context"] = {**self.class_context, **frame["@context"]}

        framed = jsonld.frame(
            {
                "@graph": response_graph,
                "@context": response_context
            }, frame, {"embed": (f"{EMBED_OPTION}")})

        if '@graph' in framed:
            return framed['@graph']
        else:
            if '@context' in framed:
                del framed['@context']
            return framed

    # UPDATE METHODS

    def traverse_obj(self, body, username):
        '''
        Utils method to insert nested resources
        Parameters
        ----------
        body : JSON with the resource (and nested resources) to insert
        username : named graph where to insert the resource

        Returns
        -------

        '''
        for key, value in body.items():
            if key != "openapi_types" and key != "attribute_map":
                if isinstance(value, list):
                    for inner_values in value:
                        if not (isinstance(inner_values, primitives.__args__)
                                or isinstance(inner_values, dict)):
                            list_of_obj = self.get_all_complex_objects(
                                inner_values)
                            if len(list_of_obj) != 0:
                                self.traverse_obj(inner_values, username)

                            if inner_values.id is None:
                                inner_values.id = generate_new_id()
                                self.insert_all_resources(
                                    inner_values, username)
                elif isinstance(value, dict):
                    pass

    def get_all_complex_objects(self, body: dict) -> list:
        """Get all the complex objects in a dictionary (recursive)
        A complex object is an object that has an id and a type

        Args:
            body (dict): The dictionary to be traversed

        Returns:
            list: a list of complex objects
        """
        l = []
        for key, value in body.__dict__.items():
            if key != "openapi_types" and key != "attribute_map":
                if isinstance(value, list):
                    for inner_values in value:
                        if not isinstance(inner_values,
                                          str) and not isinstance(
                                              inner_values, dict):
                            l.append(inner_values)
                elif isinstance(value, dict):
                    pass
        return l

    def json_to_jsonld(self, resource: dict) -> dict:
        """Convert a JSON to JSON-LD (recursive). Used by POST and PUT

        Args:
            resource (dict): The resource to be converted

        Returns:
            dict: The resource in JSON-LD
        """
        resource_dict = resource
        resource_dict["id"] = self.build_instance_uri(resource_dict["id"])
        resource_dict['@context'] = self.context
        resource_json = json.dumps(resource_dict)
        return resource_json

    # RUN QUERY METHODS

    def run_query_insert(self, request_args: dict):
        """Generate the insert query

        Args:
            request_args (dict): The request arguments (from the request)

        """
        query_string = f'{request_args["prefixes"]}' \
            f'INSERT DATA {{ GRAPH <{request_args["g"]}> ' \
            f'{{ {request_args["triples"]} }} }}'
        try:
            self.sparql.update(query_string)
        except Exception as err:
            logger.error("Exception occurred", exc_info=True)
            logger.error(query_string)
            raise err

    def run_query_delete(self, request_args: str):
        """Delete a resource

        Args:
            request_args (str): The request arguments (from the request)

        Returns:
            [type]: A tuple (message, http_code, response)
        """
        query_string = f'' \
            f'DELETE WHERE {{ GRAPH <{request_args["g"]}> ' \
            f'{{ <{request_args["resource"]}> ?p ?o . }} }}'
        try:
            logger.info("deleting %s", request_args["resource"])
            logger.debug("deleting: %s", query_string)
            self.sparql.update(query_string)
        except Exception as exception:
            logger.error("Exception occurred", exc_info=True)
            raise HTTPException(status_code=500, detail=str(exception)) from exception

        if request_args["delete_incoming_relations"]:
            query_string_reverse = f'' \
                f'DELETE WHERE {{ GRAPH <{request_args["g"]}> ' \
                f'{{ ?s ?p <{request_args["resource"]}>  }} }}'
            try:
                logger.info("deleting incoming relations %s", request_args["resource"])
                logger.debug("deleting: %s", query_string_reverse)
                self.sparql.update(query_string_reverse)
            except Exception as exception:
                logger.error("Exception occurred", exc_info=True)
                raise HTTPException(
                    status_code=500, detail=str(exception)) from exception

    def build_query_insert(self, resource_json: str):
        """Convert the JSON-LD to triple to be inserted"""
        prefixes = []
        triples = []
        g = Graph().parse(data=resource_json,
                          format='json-ld',
                          publicID=self.uri_prefix)
        s = g.serialize(format='turtle')
        for n in g.namespace_manager.namespaces():
            prefixes.append(f'PREFIX {n[0]}: <{n[1]}>')

        for line in s.split('\n'):
            if not line.startswith('@prefix'):
                triples.append(line)
        return prefixes, triples

    # UTILS METHODS
    # TODO: Refactoring as a utils method. Remove self param
    def convert_snake_dict(self, temp_context: Dict):
        """
        The Python server generated by OBA uses snake_case format for the JSON key.
        This method generates the context used on the JSON-LD operations
        Parameters
        ----------
        temp_context: Temporal dictionary with the JSON-LD context

        Returns
        -------

        """

        for key, value in temp_context.items():
            key_snake = convert_snake(key)
            self.context[key] = value
            if key_snake != key:
                self.context[key_snake] = value

    # TODO: Refactoring as a utils method. Remove self param
    def generate_graph(self, username: str) -> str:
        """Generate the graph URI

        Args:
            username (str): The username

        Returns:
            str: The graph URI
        """
        return "{}{}".format(self.named_graph_base, username)

    def build_instance_uri(self, _id: str) -> str:
        """Create the instance URI from the id and the URI prefix

        Args:
            _id (str): id of the resource

        Returns:
            str: The instance URI
        """
        if validators.url(_id):
            return _id
        return f"{self.uri_prefix}{_id}"

    @staticmethod
    def read_context(context_file):
        """
        Read the context file
        : param context_file: Absolute path of the file
        : type context_file: string
        : return: Contents of the file
        : rtype: string
        """
        try:
            with open(context_file, 'r') as reader:
                return reader.read()
        except FileNotFoundError as e:
            logging.error(f"{context_file} missing")
            raise e

    @staticmethod
    def read_template(owl_class_dir):
        queries = {}
        for file_query in os.listdir(owl_class_dir):
            filename, file_extension = os.path.splitext(file_query)
            if file_extension == ".rq":
                with open(owl_class_dir / file_query, 'r') as reader:
                    key_name = filename
                    queries[key_name] = reader.read()
        return queries

    def insert_all_resources(self, body, username):
        body_json = self.json_to_jsonld(body)
        prefixes, triples = self.build_query_insert(body_json)
        prefixes = '\n'.join(prefixes)
        triples = '\n'.join(triples)
        request_args: Dict[str, str] = {
            "prefixes": prefixes,
            "triples": triples,
            "g": self.generate_graph(username)
        }
        self.run_query_insert(request_args=request_args)

    @staticmethod
    def parse_request_arguments(**kwargs) -> Tuple[str, str, str]:
        """Parse the request arguments into variables

        Returns:
            Tuple[str, str, str]: [description]
        """
        if USERNAME_KEY in kwargs:
            username = kwargs[USERNAME_KEY]
        else:
            username = None
        owl_class_name = kwargs["rdf_type_name"]
        resource_type_uri = kwargs["rdf_type_uri"]
        return owl_class_name, resource_type_uri, username

    def dispatch_sparql_query(self, raw_sparql_query: str, request_args: str):
        """Replace the variables in the query with the request arguments and send it

        Args:
            raw_sparql_query (str): the raw query
            request_args (str): the request arguments to be replaced

        Raises:
            e: [description]

        Returns:
            dict: JSON-LD response
        """

        query_metadata = gquery.get_metadata(raw_sparql_query, self.endpoint)
        rewritten_query = query_metadata['query']
        # Rewrite query using parameter values
        if query_metadata['type'] == 'SelectQuery' or query_metadata[
                'type'] == 'ConstructQuery':
            try:
                rewritten_query = gquery.rewrite_query(
                    query_metadata['original_query'],
                    query_metadata['parameters'], request_args)
            except Exception as exception:
                logger.error("Parameters expected: %s",  query_metadata['parameters'])
                logger.error("Parameters given %s ", request_args)
                raise exception

        if PER_PAGE_KEY in request_args and "offset" in request_args:
            rewritten_query = rewritten_query.replace(
                "LIMIT 100", "LIMIT {}".format(request_args[PER_PAGE_KEY]))
            rewritten_query = rewritten_query.replace(
                "OFFSET 0", "OFFSET {}".format(request_args["offset"]))
        logger.info(rewritten_query)
        return self.sparql.query(rewritten_query)



import logging
from typing import Optional, TYPE_CHECKING, Tuple
from urllib.request import urlopen, Request
from urllib.parse import urlencode
from urllib.error import HTTPError
import base64
from rdflib import BNode

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    import typing_extensions as te


class SPARQLConnectorException(Exception):
    pass


# TODO: Pull in these from the result implementation plugins?
_response_mime_types = {
    "xml": "application/sparql-results+xml, application/rdf+xml",
    "json": "application/sparql-results+json",
    "csv": "text/csv",
    "tsv": "text/tab-separated-values",
    "application/rdf+xml": "application/rdf+xml",
    "json-ld": "application/ld+json"
}


class SPARQLConnector(object):
    """
    this class deals with nitty gritty details of talking to a SPARQL server
    """

    def __init__(
        self,
        query_endpoint: Optional[str] = None,
        update_endpoint: Optional[str] = None,
        returnFormat: str = "xml",
        method: "te.Literal['GET', 'POST', 'POST_FORM']" = "GET",
        auth: Optional[Tuple[str, str]] = None,
        **kwargs,
    ):
        """
        auth, if present, must be a tuple of (username, password) used for Basic Authentication

        Any additional keyword arguments will be passed to to the request, and can be used to setup timesouts etc.
        """

        self.returnFormat = returnFormat
        self.query_endpoint = query_endpoint
        self.update_endpoint = update_endpoint
        self.kwargs = kwargs
        self.method = method
        if auth is not None:
            if type(auth) != tuple:
                raise SPARQLConnectorException("auth must be a tuple")
            if len(auth) != 2:
                raise SPARQLConnectorException(
                    "auth must be a tuple (user, password)")
            base64string = base64.b64encode(bytes("%s:%s" % auth, "ascii"))
            self.kwargs.setdefault("headers", {})
            self.kwargs["headers"].update(
                {"Authorization": "Basic %s" % base64string.decode("utf-8")})

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        if method not in ("GET", "POST", "POST_FORM"):
            raise SPARQLConnectorException(
                'Method must be "GET", "POST", or "POST_FORM"')

        self._method = method

    def query(self, query, default_graph: str = None, named_graph: str = None):

        if not self.query_endpoint:
            raise SPARQLConnectorException("Query endpoint not set!")

        params = {}
        # this test ensures we don't have a useless (BNode) default graph URI, which calls to Graph().query() will add
        if default_graph is not None and type(default_graph) != BNode:
            params["default-graph-uri"] = default_graph

        headers = {"Accept": _response_mime_types[self.returnFormat]}

        args = dict(self.kwargs)

        # merge params/headers dicts
        args.setdefault("params", {})

        args.setdefault("headers", {})
        args["headers"].update(headers)

        if self.method == "GET":
            params["query"] = query
            args["params"].update(params)
            qsa = "?" + urlencode(args["params"])
            try:
                res = urlopen(
                    Request(self.query_endpoint + qsa,
                            headers=args["headers"]))
            except Exception as e:
                raise ValueError(
                    "You did something wrong formulating either the URI or your SPARQL query"
                )
        elif self.method == "POST":
            args["headers"].update(
                {"Content-Type": "application/sparql-query"})
            qsa = "?" + urlencode(params)
            try:
                res = urlopen(
                    Request(
                        self.query_endpoint + qsa,
                        data=query.encode(),
                        headers=args["headers"],
                    ))
            except HTTPError as e:
                return e.code, str(e), None
        elif self.method == "POST_FORM":
            params["query"] = query
            args["params"].update(params)
            try:
                res = urlopen(
                    Request(
                        self.query_endpoint,
                        data=urlencode(args["params"]).encode(),
                        headers=args["headers"],
                    ))
            except HTTPError as e:
                return e.code, str(e), None
        else:
            raise SPARQLConnectorException("Unknown method %s" % self.method)
        return res.read()

    def update(
        self,
        query,
        default_graph: Optional[str] = None,
        named_graph: Optional[str] = None,
    ):
        """Method to send a SPARQL update query to the endpoint.

        Args:
            query (str): the query to send to the endpoint
            default_graph (Optional[str], optional): The default named graph. Defaults to None.
            named_graph (Optional[str], optional): The named graph. Defaults to None.

        Raises:
            SPARQLConnectorException: Connection error
        """
        if not self.update_endpoint:
            raise SPARQLConnectorException("Query endpoint not set!")

        params = {}

        if default_graph is not None:
            params["using-graph-uri"] = default_graph

        if named_graph is not None:
            params["using-named-graph-uri"] = named_graph

        headers = {
            "Accept": _response_mime_types[self.returnFormat],
            "Content-Type": "application/sparql-update",
        }

        args = dict(self.kwargs)  # other QSAs

        args.setdefault("params", {})
        args["params"].update(params)
        args.setdefault("headers", {})
        args["headers"].update(headers)

        qsa = "?" + urlencode(args["params"])
        res = urlopen(
            Request(self.update_endpoint + qsa,
                    data=query.encode(),
                    headers=args["headers"]))



from .query_manager import QueryManager


def init(**kwargs):
    """Initialize OBA Sparql API Client.
    Keyword Args:
        queries_dir (str): Directory of queries.
        context_dir (str): Directory of context.
        queries_types (array): Queries types enabled.
    """
    return QueryManager(**kwargs)



import pathlib
from setuptools import setup
import codecs

version = ""
with open('VERSION') as fp:
    version = fp.read()


# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()


with codecs.open('README.md', mode='r', encoding='utf-8') as f:
    long_description = f.read()

# This call to setup() does all the work
setup(
    name="obasparql",
    version=version,
    description="OBA Sparql Manager",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/KnowledgeCaptureAndDiscovery/oba_sparql",
    author="Maximiliano Osorio",
    author_email="mosorio@isi.edu",
    license="Apache-2",
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Science/Research",
        "Operating System :: Unix",
    ],
    packages=["obasparql"],
    include_package_data=True,
    install_requires=[
        "validators==0.20.0",
        "pythonql3==0.9.83",
        "pyaml==21.10.1",
        "rdflib==6.2.0",
        "requests>=2.20.0",
        "simplejson==3.17.6",
        "six>=1.11.0",
        "urllib3<1.27,>=1.21.1 ",
        "webencodings==0.5.1",
        "starlette==0.20.4",
        "PyLD>=2.0.3",
    ]
)
